// Copyright 2023 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package bench

import (
	"testing"

	"github.com/google/cel-go/cel"
)

func BenchmarkReferenceCases(b *testing.B) {
	stdenv, err := cel.NewEnv()
	if err != nil {
		b.Fatalf("cel.NewEnv() failed: %v", err)
	}
	RunReferenceCases(b, stdenv)
}

func BenchmarkReferenceCheckerCases(b *testing.B) {
	RunReferenceDynamicEnvCases(b)
}
