// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.28.1
// 	protoc        v3.21.5
// source: test/proto2pb/test_extensions.proto

package proto2pb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	wrapperspb "google.golang.org/protobuf/types/known/wrapperspb"
	reflect "reflect"
	sync "sync"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type ExternalMessageType struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields
}

func (x *ExternalMessageType) Reset() {
	*x = ExternalMessageType{}
	if protoimpl.UnsafeEnabled {
		mi := &file_test_proto2pb_test_extensions_proto_msgTypes[0]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *ExternalMessageType) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ExternalMessageType) ProtoMessage() {}

func (x *ExternalMessageType) ProtoReflect() protoreflect.Message {
	mi := &file_test_proto2pb_test_extensions_proto_msgTypes[0]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ExternalMessageType.ProtoReflect.Descriptor instead.
func (*ExternalMessageType) Descriptor() ([]byte, []int) {
	return file_test_proto2pb_test_extensions_proto_rawDescGZIP(), []int{0}
}

var file_test_proto2pb_test_extensions_proto_extTypes = []protoimpl.ExtensionInfo{
	{
		ExtendedType:  (*ExampleType)(nil),
		ExtensionType: (*ExampleType)(nil),
		Field:         100,
		Name:          "google.expr.proto2.test.nested_example",
		Tag:           "bytes,100,opt,name=nested_example",
		Filename:      "test/proto2pb/test_extensions.proto",
	},
	{
		ExtendedType:  (*ExampleType)(nil),
		ExtensionType: (*int32)(nil),
		Field:         101,
		Name:          "google.expr.proto2.test.int32_ext",
		Tag:           "varint,101,opt,name=int32_ext",
		Filename:      "test/proto2pb/test_extensions.proto",
	},
	{
		ExtendedType:  (*ExampleType)(nil),
		ExtensionType: (*wrapperspb.Int32Value)(nil),
		Field:         102,
		Name:          "google.expr.proto2.test.int32_wrapper_ext",
		Tag:           "bytes,102,opt,name=int32_wrapper_ext",
		Filename:      "test/proto2pb/test_extensions.proto",
	},
	{
		ExtendedType:  (*ExampleType)(nil),
		ExtensionType: (*int64)(nil),
		Field:         201,
		Name:          "google.expr.proto2.test.ExternalMessageType.int64_ext",
		Tag:           "varint,201,opt,name=int64_ext",
		Filename:      "test/proto2pb/test_extensions.proto",
	},
}

// Extension fields to ExampleType.
var (
	// optional google.expr.proto2.test.ExampleType nested_example = 100;
	E_NestedExample = &file_test_proto2pb_test_extensions_proto_extTypes[0]
	// optional int32 int32_ext = 101;
	E_Int32Ext = &file_test_proto2pb_test_extensions_proto_extTypes[1]
	// optional google.protobuf.Int32Value int32_wrapper_ext = 102;
	E_Int32WrapperExt = &file_test_proto2pb_test_extensions_proto_extTypes[2]
	// optional int64 int64_ext = 201;
	E_ExternalMessageType_Int64Ext = &file_test_proto2pb_test_extensions_proto_extTypes[3]
)

var File_test_proto2pb_test_extensions_proto protoreflect.FileDescriptor

var file_test_proto2pb_test_extensions_proto_rawDesc = []byte{
	0x0a, 0x23, 0x74, 0x65, 0x73, 0x74, 0x2f, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x32, 0x70, 0x62, 0x2f,
	0x74, 0x65, 0x73, 0x74, 0x5f, 0x65, 0x78, 0x74, 0x65, 0x6e, 0x73, 0x69, 0x6f, 0x6e, 0x73, 0x2e,
	0x70, 0x72, 0x6f, 0x74, 0x6f, 0x12, 0x17, 0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x2e, 0x65, 0x78,
	0x70, 0x72, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x32, 0x2e, 0x74, 0x65, 0x73, 0x74, 0x1a, 0x1e,
	0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x2f, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x62, 0x75, 0x66, 0x2f,
	0x77, 0x72, 0x61, 0x70, 0x70, 0x65, 0x72, 0x73, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x1a, 0x22,
	0x74, 0x65, 0x73, 0x74, 0x2f, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x32, 0x70, 0x62, 0x2f, 0x74, 0x65,
	0x73, 0x74, 0x5f, 0x61, 0x6c, 0x6c, 0x5f, 0x74, 0x79, 0x70, 0x65, 0x73, 0x2e, 0x70, 0x72, 0x6f,
	0x74, 0x6f, 0x22, 0x59, 0x0a, 0x13, 0x45, 0x78, 0x74, 0x65, 0x72, 0x6e, 0x61, 0x6c, 0x4d, 0x65,
	0x73, 0x73, 0x61, 0x67, 0x65, 0x54, 0x79, 0x70, 0x65, 0x32, 0x42, 0x0a, 0x09, 0x69, 0x6e, 0x74,
	0x36, 0x34, 0x5f, 0x65, 0x78, 0x74, 0x12, 0x24, 0x2e, 0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x2e,
	0x65, 0x78, 0x70, 0x72, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x32, 0x2e, 0x74, 0x65, 0x73, 0x74,
	0x2e, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x54, 0x79, 0x70, 0x65, 0x18, 0xc9, 0x01, 0x20,
	0x01, 0x28, 0x03, 0x52, 0x08, 0x69, 0x6e, 0x74, 0x36, 0x34, 0x45, 0x78, 0x74, 0x3a, 0x71, 0x0a,
	0x0e, 0x6e, 0x65, 0x73, 0x74, 0x65, 0x64, 0x5f, 0x65, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x12,
	0x24, 0x2e, 0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x2e, 0x65, 0x78, 0x70, 0x72, 0x2e, 0x70, 0x72,
	0x6f, 0x74, 0x6f, 0x32, 0x2e, 0x74, 0x65, 0x73, 0x74, 0x2e, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c,
	0x65, 0x54, 0x79, 0x70, 0x65, 0x18, 0x64, 0x20, 0x01, 0x28, 0x0b, 0x32, 0x24, 0x2e, 0x67, 0x6f,
	0x6f, 0x67, 0x6c, 0x65, 0x2e, 0x65, 0x78, 0x70, 0x72, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x32,
	0x2e, 0x74, 0x65, 0x73, 0x74, 0x2e, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x54, 0x79, 0x70,
	0x65, 0x52, 0x0d, 0x6e, 0x65, 0x73, 0x74, 0x65, 0x64, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65,
	0x3a, 0x41, 0x0a, 0x09, 0x69, 0x6e, 0x74, 0x33, 0x32, 0x5f, 0x65, 0x78, 0x74, 0x12, 0x24, 0x2e,
	0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x2e, 0x65, 0x78, 0x70, 0x72, 0x2e, 0x70, 0x72, 0x6f, 0x74,
	0x6f, 0x32, 0x2e, 0x74, 0x65, 0x73, 0x74, 0x2e, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x54,
	0x79, 0x70, 0x65, 0x18, 0x65, 0x20, 0x01, 0x28, 0x05, 0x52, 0x08, 0x69, 0x6e, 0x74, 0x33, 0x32,
	0x45, 0x78, 0x74, 0x3a, 0x6d, 0x0a, 0x11, 0x69, 0x6e, 0x74, 0x33, 0x32, 0x5f, 0x77, 0x72, 0x61,
	0x70, 0x70, 0x65, 0x72, 0x5f, 0x65, 0x78, 0x74, 0x12, 0x24, 0x2e, 0x67, 0x6f, 0x6f, 0x67, 0x6c,
	0x65, 0x2e, 0x65, 0x78, 0x70, 0x72, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x32, 0x2e, 0x74, 0x65,
	0x73, 0x74, 0x2e, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x54, 0x79, 0x70, 0x65, 0x18, 0x66,
	0x20, 0x01, 0x28, 0x0b, 0x32, 0x1b, 0x2e, 0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x2e, 0x70, 0x72,
	0x6f, 0x74, 0x6f, 0x62, 0x75, 0x66, 0x2e, 0x49, 0x6e, 0x74, 0x33, 0x32, 0x56, 0x61, 0x6c, 0x75,
	0x65, 0x52, 0x0f, 0x69, 0x6e, 0x74, 0x33, 0x32, 0x57, 0x72, 0x61, 0x70, 0x70, 0x65, 0x72, 0x45,
	0x78, 0x74, 0x42, 0x28, 0x5a, 0x26, 0x67, 0x69, 0x74, 0x68, 0x75, 0x62, 0x2e, 0x63, 0x6f, 0x6d,
	0x2f, 0x67, 0x6f, 0x6f, 0x67, 0x6c, 0x65, 0x2f, 0x63, 0x65, 0x6c, 0x2d, 0x67, 0x6f, 0x2f, 0x74,
	0x65, 0x73, 0x74, 0x2f, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x32, 0x70, 0x62,
}

var (
	file_test_proto2pb_test_extensions_proto_rawDescOnce sync.Once
	file_test_proto2pb_test_extensions_proto_rawDescData = file_test_proto2pb_test_extensions_proto_rawDesc
)

func file_test_proto2pb_test_extensions_proto_rawDescGZIP() []byte {
	file_test_proto2pb_test_extensions_proto_rawDescOnce.Do(func() {
		file_test_proto2pb_test_extensions_proto_rawDescData = protoimpl.X.CompressGZIP(file_test_proto2pb_test_extensions_proto_rawDescData)
	})
	return file_test_proto2pb_test_extensions_proto_rawDescData
}

var file_test_proto2pb_test_extensions_proto_msgTypes = make([]protoimpl.MessageInfo, 1)
var file_test_proto2pb_test_extensions_proto_goTypes = []interface{}{
	(*ExternalMessageType)(nil),   // 0: google.expr.proto2.test.ExternalMessageType
	(*ExampleType)(nil),           // 1: google.expr.proto2.test.ExampleType
	(*wrapperspb.Int32Value)(nil), // 2: google.protobuf.Int32Value
}
var file_test_proto2pb_test_extensions_proto_depIdxs = []int32{
	1, // 0: google.expr.proto2.test.nested_example:extendee -> google.expr.proto2.test.ExampleType
	1, // 1: google.expr.proto2.test.int32_ext:extendee -> google.expr.proto2.test.ExampleType
	1, // 2: google.expr.proto2.test.int32_wrapper_ext:extendee -> google.expr.proto2.test.ExampleType
	1, // 3: google.expr.proto2.test.ExternalMessageType.int64_ext:extendee -> google.expr.proto2.test.ExampleType
	1, // 4: google.expr.proto2.test.nested_example:type_name -> google.expr.proto2.test.ExampleType
	2, // 5: google.expr.proto2.test.int32_wrapper_ext:type_name -> google.protobuf.Int32Value
	6, // [6:6] is the sub-list for method output_type
	6, // [6:6] is the sub-list for method input_type
	4, // [4:6] is the sub-list for extension type_name
	0, // [0:4] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_test_proto2pb_test_extensions_proto_init() }
func file_test_proto2pb_test_extensions_proto_init() {
	if File_test_proto2pb_test_extensions_proto != nil {
		return
	}
	file_test_proto2pb_test_all_types_proto_init()
	if !protoimpl.UnsafeEnabled {
		file_test_proto2pb_test_extensions_proto_msgTypes[0].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*ExternalMessageType); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: file_test_proto2pb_test_extensions_proto_rawDesc,
			NumEnums:      0,
			NumMessages:   1,
			NumExtensions: 4,
			NumServices:   0,
		},
		GoTypes:           file_test_proto2pb_test_extensions_proto_goTypes,
		DependencyIndexes: file_test_proto2pb_test_extensions_proto_depIdxs,
		MessageInfos:      file_test_proto2pb_test_extensions_proto_msgTypes,
		ExtensionInfos:    file_test_proto2pb_test_extensions_proto_extTypes,
	}.Build()
	File_test_proto2pb_test_extensions_proto = out.File
	file_test_proto2pb_test_extensions_proto_rawDesc = nil
	file_test_proto2pb_test_extensions_proto_goTypes = nil
	file_test_proto2pb_test_extensions_proto_depIdxs = nil
}
